#include <stdio.h>
#include <pthread.h>

union ABAvalue
{
   volatile long long llvalue;
   struct element
   {
     volatile int version;
     volatile int value;
   } nested;
};

union ABAvalue buffer[16];
volatile int addhere;
volatile int removehere;
int counter = 0;

#ifdef __sparc
long long CAS( volatile long long* addr, long long ov, long long nv)
{
  asm volatile( "casx %1, %2, %0": 
                "=r"(nv): 
                "m"(*addr),"r"(ov),"0"(nv):
                "memory" );
  return nv;
}
#else
long long CAS( volatile long long * addr, long long ov, long long nv )
{
  asm volatile( "lock; cmpxchg %2, %1":
                "=a"(ov):
                "m"(*addr),"r"(nv),"a"(ov):
                "memory" );
  return ov;
}
#endif

void clearbuffer()
{
  addhere = 0;
  removehere = 0;
  for( int i=0; i<16; i++ ) 
  { 
    buffer[i].llvalue = 0;
  }
}

int next( int current )
{
  return ( current + 1 ) & 15;
}

int nextupdate( int current )
{
  if ( current == 15 ) { counter++; }  
  return ( current + 1 ) & 15 ;
}

void addtobuffer( int value )
{
  int success = 0;
  union ABAvalue current;
  while ( !success )
  {
    current = buffer[ next(addhere) ];
    if ( current.nested.value == 0 )
    {
      union ABAvalue nextvalue;
      nextvalue.nested.version = counter;
      nextvalue.nested.value = value;
      if ( CAS( &buffer[ next(addhere) ].llvalue, 
                 current.llvalue, 
                 nextvalue.llvalue )
           == current.llvalue )
      {
        addhere = nextupdate(addhere);
        success = 1;
      }
    }
  }
}

int removefrombuffer()
{
  union ABAvalue current;
  int success = 0;
  int value;
  while ( !success )
  {
    current = buffer[next(removehere)];
    if ( current.nested.value != 0 )
    {
      union ABAvalue nextvalue;
      value = current.nested.value;
      nextvalue.nested.version = counter;
      nextvalue.nested.value = 0;
      if ( CAS( &buffer[next(removehere)].llvalue,
                 current.llvalue,
                 nextvalue.llvalue )  
           == current.llvalue )
      {
        removehere = next(removehere);
        success = 1;
      }
    }
  }
  return value;
}

void * producer( void *param )
{
  for( int i=1; i<10000000; i++ )
  {
    addtobuffer( i );
  }
}

void * consumer ( void *param )
{
  while ( removefrombuffer() != 9999999 ) {}
}

int main()
{
  clearbuffer();
  pthread_t threads[2];
  pthread_create( &threads[0], 0, producer, 0 );
  pthread_create( &threads[1], 0, consumer, 0 );
  pthread_join( threads[1], 0 );
  pthread_join( threads[0], 0 );
}

