#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#define SIZE 128*1024*1024
int** memory;

double now()
{
  struct timeval time;
  gettimeofday( &time, 0 );
  return (double)time.tv_sec + (double)time.tv_usec / 1000000.0;
}

#ifdef __sun
#include <sys/processor.h>
#include <sys/procset.h>

int threadbind( int proc )
{
  return processor_bind( P_LWPID, P_MYID, proc, 0 );
}
#endif

#ifdef linux
#include <sched.h>
#include <sys/syscall.h>

pid_t gettid()
{
  return syscall( __NR_gettid );
}

int threadbind( int proc )
{
  cpu_set_t cpuset;
  CPU_ZERO( &cpuset );
  CPU_SET( proc, &cpuset );
  return sched_setaffinity( gettid(), sizeof(cpu_set_t), &cpuset );
}
#endif

void setup()
{
  int i, j;
  threadbind( 0 );
  memory = (int**)calloc( sizeof(int*), SIZE+1 );
  for( j=0; j<16; j++ ) 
  {
    for( i=16+j; i<=SIZE; i++) 
    { 
      memory[i] = (int*)&memory[ i-16 ]; 
    }
    memory[j] = (int*)&memory[ SIZE-16+j+1 ];
  }
}

void *use( void* param )
{
  if ( threadbind( (int)param ) == 0 ) 
  {
    int* next;
    int i;
    double start = now();
    next = memory[0];
    for( i=0; i<SIZE; i++ ) 
    { 
      next = (int*)*next; 
    }
    if ( (int)next == 1 ) { printf( "a" ); }
    double end = now();
    printf( "Time %14.2f ns Binding %i/s\n",
            (end - start) * 1000000000.0 / ((double)SIZE), (int)param );
  } else { perror( "Error" ); }
}

int main( int argc, char* argv[] )
{
  pthread_t thread;
  int nthreads = 1;
  int i;
  if ( argc > 1 ) { nthreads = atoi( argv[1] ); }
  setup();
  for( i=0; i<nthreads; i++ ) 
  {
    pthread_create( &thread, 0, use, (void*)i );
    pthread_join( thread, 0 );
  }
  return 0;
}
